
class CommonAlgs:

    @staticmethod
    def split_nth(n: int, line: str or bytes) -> list:
        return [line[i:i + n] for i in range(0, len(line), n)]

    @staticmethod
    def repeated_key_xor(plain_text: bytes, key: bytes) -> bytes:
        pt = plain_text
        len_key = len(key)
        encoded = []

        for i in range(0, len(pt)):
            encoded.append(pt[i] ^ key[i % len_key])
        return bytes(encoded)






import sys
from .CHAF import *
import secrets
from .Modes import *
from .CommonAlgs import CommonAlgs
from .Padding import PKCS7
class Piranha:
    ECB = Modes.ECB
    CBC = Modes.CBC
    CTR = Modes.CTR
    EAA = Modes.EAA
    BlockSize = 64
    _uses_IV = [CTR, CBC]



    def __init__(self, key, mode: int, *args, **kwargs):
        self.mode = Modes.new(key, mode, *args, **kwargs)
        self.key = key
        self.m = mode
        self.iv = self.mode.iv
        self.args = args
        self.kwargs = kwargs

    def update(self, data: bytes):
        self.mode.update(data)





    @staticmethod
    def pad(data: bytes, blockSize=None) -> bytes:
        if blockSize is None: blockSize = Piranha.BlockSize
        return PKCS7(blockSize).pad(data)

    @staticmethod
    def unpad(data: bytes, blockSize=None) -> bytes:
        if blockSize is None: blockSize = Piranha.BlockSize
        return PKCS7(blockSize).unpad(data)

    def encryptionFunction(self, data: bytes, key, *args, **kwargs) -> bytes:
        func = FeistelN.fRAB_with_nonce(key, rep=1, rev=1)
        return FeistelN().DE(data, 4, func, 'e', 's')

    def decryptionFunction(self, data: bytes, key, *args, **kwargs) -> bytes:
        func = FeistelN.fRAB_with_nonce(key, rep=1, rev=1)
        return FeistelN().DE(data, 4, func, 'd', 's')

    def encrypt(self, data: bytes = None) -> bytes:
        if data is None: data = self.mode.data
        return self.mode.encrypt(data, self.encryptionFunction)

    def decrypt(self, cipher: bytes) -> bytes:
        if cipher is None: cipher = self.mode.data
        return self.mode.decrypt(cipher, self.decryptionFunction)

    def HMAC(self, data: bytes) -> bytes:
        return self.mode.HMAC(data)

    def verify(self, data: bytes, mac: bytes) -> bytes:
        return self.mode.verify(data, mac)





from __future__ import annotations

import hmac
import secrets
from .CommonAlgs import CommonAlgs
from .Padding import PKCS7
class Modes:
    ECB = 0
    CBC = 1
    CTR = 2
    EAA = 3
    _uses_IV = [CTR, CBC]
    BlockSize = 64

    _registry = {}


    @staticmethod
    def new(key: bytes = None, prefix:int=0, *args, **kwargs) -> Modes:
        return Modes(key, prefix, *args, **kwargs)

    def __init_subclass__(cls, prefix, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._registry[prefix] = cls

    def __new__(cls, key=None, prefix=0, *args, **kwargs):
        subclass = cls._registry[prefix]
        obj = object.__new__(subclass)
        obj.key = key
        obj.mode = prefix
        obj.iv = secrets.token_bytes(16)
        if 'iv' in kwargs.keys():
            obj.iv = kwargs['iv']
        obj.BlockSize = 64
        if 'BlockSize' in kwargs.keys():
            obj.BlockSize = kwargs['BlockSize']
        obj.data = b''
        if 'data' in kwargs.keys():
            obj.data = PKCS7(obj.BlockSize).pad(kwargs['data'])
        obj.args = args
        obj.kwargs = kwargs
        obj.key = key
        return obj


    @staticmethod
    def repeated_key_xor(plain_text, key) -> bytes:
        return CommonAlgs.repeated_key_xor(plain_text, key)

    def HMAC(self, data: bytes) -> bytes:
        return hmac.new(self.key + self.iv, data, digestmod="sha512").digest()

    def verify(self, data: bytes, mac: bytes) -> bytes:
        return hmac.compare_digest(hmac.new(self.key + self.iv, data, digestmod="sha512").digest(), mac)

    @staticmethod
    def split_nth(n: int, line: str or bytes) -> str or bytes:
        return [line[i:i + n] for i in range(0, len(line), n)]


    def pad(self, data: bytes) -> bytes:
        return PKCS7(self.BlockSize).pad(data)

    @staticmethod
    def unpad(self, data: bytes) -> bytes:
        return PKCS7(self.BlockSize).unpad(data)

    def encrypt(self, data: bytes, func, *args, **kwargs) -> bytes:
        raise NotImplementedError

    def update(self, data: bytes)  -> None:
        self.data = data

    def decrypt(self, cipher: bytes, func, *args, **kwargs) -> bytes:
        raise NotImplementedError

class ModesCTR(Modes, prefix=Modes.CTR):
    def encrypt(self, data: bytes, func, *args, **kwargs) -> bytes:
        if data is None: data = self.data
        data = self.pad(data)
        repUnit = 16
        if 'n' in kwargs:
            repUnit = kwargs['n']
        dataList = self.split_nth(self.BlockSize, data)
        times = len(dataList)
        if times < repUnit: repUnit = times
        encryptedIVs = []
        for i in range(1, repUnit+1):
            nonce = self.iv + bytes([i])
            iv = func(nonce, self.key, *args, **kwargs)
            encryptedIVs.append(iv)
        out = [self.repeated_key_xor(self.repeated_key_xor(encryptedIVs[i % len(encryptedIVs)], c), self.key)
               for i, c in enumerate(dataList)]
        return b''.join(out)

    def decrypt(self, cipher: bytes, func, *args, **kwargs) -> bytes:
        return self.encrypt(cipher, func, *args, **kwargs)

class ModesCBC(Modes, prefix=Modes.CBC):
    def encrypt(self, data: bytes, func, *args, **kwargs) -> bytes:
        if data is None: data = self.data
        dataList = self.split_nth(self.BlockSize, data)
        times = len(dataList)
        encrypted = []
        nextXOR = self.iv
        for i in range(times):
            xoredData = self.repeated_key_xor(dataList[i], nextXOR)
            encryptedData = func(xoredData, self.key, *args, **kwargs)
            encrypted.append(encryptedData)
            nextXOR = encryptedData
        return self.repeated_key_xor(b''.join(encrypted), self.key)

    def decrypt(self, cipher: bytes, func, *args, **kwargs) -> bytes:
        cipher = self.repeated_key_xor(cipher, self.key)
        dataList = Modes.split_nth(self.BlockSize, cipher)
        times = len(dataList)
        decrypted = []
        nextXOR = self.iv
        for i in range(times):
            decryptedData = func(dataList[i], self.key, *args, **kwargs)
            xored = Modes.repeated_key_xor(decryptedData, nextXOR)
            decrypted.append(xored)
            nextXOR = dataList[i]
        return b''.join(decrypted)

class ModesECB(Modes, prefix=Modes.ECB):
    def encrypt(self, data: bytes, func, *args, **kwargs) -> bytes:
        ra = []
        ml = Modes.split_nth(self.BlockSize, data)
        for i in ml:
            ra.append(func(i, self.key, *args, **kwargs))
        return b"".join(ra)

    def decrypt(self, cipher: bytes, func, *args, **kwargs) -> bytes:
        ra1 = []
        message = Modes.split_nth(self.BlockSize, cipher)
        for e in message:
            ra1.append(func(e, self.key, *args, **kwargs))
        return b''.join(ra1)


class ModesEAA(Modes, prefix=Modes.EAA):

    def encrypt(self, data: bytes, func, *args, **kwargs) -> bytes:
        cipher = Modes(self.key, Modes.CTR, data=data, iv=self.iv)
        d = cipher.encrypt(data, func)
        hmac = cipher.HMAC(data)
        self.iv = secrets.token_bytes(16)
        return hmac + cipher.iv + d

    def decrypt(self, cipher: bytes, func, *args, **kwargs) -> bytes:
        hmac = cipher[:64]
        iv = cipher[64:80]
        data = cipher[80:]
        cipher = Modes(self.key, Modes.CTR, data=data, iv=iv)
        pt = cipher.decrypt(data, func)
        v = cipher.verify(pt, hmac)
        if not v: raise ValueError("HMACs don't match!")
        return pt


from .CHAF import *
from base64 import b64encode, b64decode
from .Piranha import Piranha

class PEM:
    @staticmethod
    def split_nth(n, line):
        return [line[i:i + n] for i in range(0, len(line), n)]

    @staticmethod
    def validate(data: bytes) -> bool:
        close_to_open = {b"----END   ": b"----BEGIN "}
        stack = []
        split = data.split(b"\n")
        for d in split:
            if not d.startswith(b"----"):
                continue
            if d[:10] in close_to_open:
                if not stack: return False
                top_element = stack.pop()
                if close_to_open[d[:10]] != top_element:
                    return False
            else:
                stack.append(d[:10])
        return len(stack) == 0



    @staticmethod
    def export_PEM(b: bytes, passcode: bytes, marker: bytes, n=64):
        if not (len(passcode) == 0 or passcode is None):
            cipher = Piranha(key=passcode, mode=Piranha.EAA)
            data = cipher.encrypt(Piranha.pad(b, Piranha.BlockSize))
            b = data
        e = b64encode(b)
        l = PEM.split_nth(n, e)
        out = b"----BEGIN " + marker + b"----\n"
        out += b"\n".join(l)
        out += b"\n----END   " + marker + b"----"

        return out

    @staticmethod
    def import_PEM(b: bytes, passcode: bytes):
        v = PEM.validate(b)
        if not v: raise ValueError("Invalid data!")
        l = b.split(b"\n")
        l.pop(0)
        l.pop(-1)
        i = b64decode(b''.join(l))
        d = i
        if not (len(passcode) == 0 or passcode is None):
            data = i
            cipher = Piranha(key=passcode, mode=Piranha.EAA)
            d = Piranha.unpad(cipher.decrypt(data))
        return d
